from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    import matplotlib.axes
    import pandas as pd

bad_tag = "[bold red]"
good_tag = "[bold green]"
cautious_tag = "[bold yellow]"


EPOCH_COLOR_MAP = {
    "RFMapping": "black",
    "DynamicRouting1": "blue",
    "OptoTagging": "cyan",
    "Spontaneous": "red",
    "SpontaneousRewards": "magenta",
}


def add_epoch_color_bars(
    ax: matplotlib.axes.Axes,
    epochs: pd.DataFrame,
    epoch_color_map: dict[str, Any] | None = None,
    **text_kwargs,
) -> None:
    if epoch_color_map is None:
        epoch_color_map = EPOCH_COLOR_MAP
    text_kwargs = {
        "y": 0,
        "ha": "center",
        "rotation": 0,
        "fontsize": 6,
        "zorder": 1,
    } | text_kwargs
    for _, epoch in epochs.iterrows():
        epoch_name = next((k for k in epoch.tags if k in epoch_color_map), "")
        color = epoch_color_map[epoch_name] if epoch_name else "black"
        ax.axvspan(epoch.start_time, epoch.stop_time, alpha=0.1, color=color, zorder=0)
        ax.text(
            x=(epoch.stop_time + epoch.start_time) / 2,
            s=epoch_name,
            **text_kwargs,
        )


def bad_string(base: str) -> str:
    return bad_tag + base + bad_tag


def good_string(base: str) -> str:
    return good_tag + base + good_tag


def cautious_string(base: str) -> str:
    return cautious_tag + base + cautious_tag


def determine_string_valence(
    value: Any, good_criterion: bool, bad_criterion: bool
) -> Callable[[str], str]:
    if good_criterion:
        return good_string

    elif bad_criterion:
        return bad_string

    else:
        return cautious_string


def add_valence_to_string(
    basestring: str, value: float | int, good_criterion: bool, bad_criterion: bool
) -> str:
    valence_func = determine_string_valence(value, good_criterion, bad_criterion)

    return valence_func(basestring)



from __future__ import annotations

import datetime
import random
from collections.abc import Iterable
from typing import TYPE_CHECKING

import cv2
import matplotlib.figure
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import rich
import upath

import npc_sessions.utils as utils

if TYPE_CHECKING:
    import npc_sessions

import npc_sessions.plots.plot_utils as plot_utils
import npc_sessions.utils as utils


def plot_video_info(
    session: npc_sessions.DynamicRoutingSession,
) -> None:
    "Not a plot: prints info to stdout"

    augmented_camera_info = utils.get_augmented_camera_info(
        session.sync_data, *session.video_paths
    )

    for camera, info in augmented_camera_info.items():
        rich.print(f"[bold]{camera} camera stats[bold]")

        frame_rate = info["FPS"]
        frame_rate_string = plot_utils.add_valence_to_string(
            f"Frame Rate: {frame_rate} \t",
            frame_rate,
            abs(frame_rate - 60) < 0.01,
            abs(frame_rate - 60) > 0.05,
        )

        lost_frame_percentage = 100 * info["FramesLostCount"] / info["FramesRecorded"]
        lost_frame_string = plot_utils.add_valence_to_string(
            f"Lost frame percentage: {np.round(lost_frame_percentage, 3)} \t",
            lost_frame_percentage,
            lost_frame_percentage < 0.01,
            lost_frame_percentage > 0.05,
        )

        frame_diff_from_expected = info["expected_minus_actual"]
        frame_diff_string = plot_utils.add_valence_to_string(
            f"Frames expected minus actual: {frame_diff_from_expected}",
            frame_diff_from_expected,
            abs(frame_diff_from_expected) < 1,
            abs(frame_diff_from_expected) > 10,
        )

        rich.print(frame_rate_string + lost_frame_string + frame_diff_string)


def plot_camera_frame_grabs_simple(
    session: npc_sessions.DynamicRoutingSession,
    paths: Iterable[upath.UPath] | None = None,
    num_frames_to_grab: int = 5,
) -> matplotlib.figure.Figure:
    """Just plots evenly spaced frames, no concept of epochs.

    video frames across cameras aren't synced .
    """
    if paths is None:
        paths = session.video_paths

    paths = tuple(paths)

    fig = plt.figure(
        figsize=[10, 3 * len(paths)], constrained_layout=True, facecolor="0.5"
    )
    gs = gridspec.GridSpec(len(paths), num_frames_to_grab, figure=fig)
    gs.update(wspace=0.0, hspace=0.0)
    for idx, video_path in enumerate(paths):
        # get frames to plot

        v = utils.get_video_data(video_path)  # TODO open with upath from cloud

        frame_delta = np.ceil(v.get(cv2.CAP_PROP_FRAME_COUNT) / num_frames_to_grab + 1)
        frames_of_interest = np.arange(
            v.get(cv2.CAP_PROP_FPS), v.get(cv2.CAP_PROP_FRAME_COUNT), frame_delta
        )

        for i, f in enumerate(frames_of_interest):
            v.set(cv2.CAP_PROP_POS_FRAMES, int(f))
            ret, frame = v.read()
            ax = fig.add_subplot(gs[idx, i])
            im = ax.imshow(frame)
            im.set_clim([0, 255])
            # ax.axis('off')
            ax.tick_params(
                top=False,
                bottom=False,
                left=False,
                right=False,
                labelleft=False,
                labelbottom=False,
            )
            ax.set_title(
                datetime.timedelta(seconds=f / v.get(cv2.CAP_PROP_FPS)), fontsize=10
            )
    return fig


def plot_video_frames_with_licks(
    session: npc_sessions.DynamicRoutingSession,
    trial_idx: int | None = None,
    lick_time: float | None = None,
):
    NUM_LICKS = 3 if (trial_idx is None and lick_time is None) else 1
    NUM_CAMERAS = 2  # 1 x face, 1 x body

    FRAMES_EITHER_SIDE_OF_LICK = 2
    """Symmetric around frame closest to lick"""
    FRAMES_PER_ROW = FRAMES_EITHER_SIDE_OF_LICK * 2 + 1

    ROWS_PER_VIDEO = 2  # 1 x camera frames, 1 x eventplots
    ROWS_PER_LICK = ROWS_PER_VIDEO * NUM_CAMERAS

    if lick_time is None:
        response_times: npt.NDArray = (
            (session.trials[:].query("is_response").response_time.to_numpy())
            if trial_idx is None
            else (session.trials[trial_idx].response_time.to_numpy())
        )
        if not response_times.shape:
            raise ValueError(f"{session.id} has no lick response times")
        lick_times = sorted(
            random.sample(tuple(response_times[~np.isnan(response_times)]), NUM_LICKS)
        )
    else:
        lick_times = [lick_time]

    fig = plt.figure(figsize=[12, 5 * NUM_LICKS], facecolor="0.5")
    # fig, axes = plt.subplots(NUM_LICKS * ROWS_PER_LICK, FRAMES_PER_ROW,)
    # constrained_layout=True,
    gs = gridspec.GridSpec(
        NUM_LICKS * ROWS_PER_LICK,
        FRAMES_PER_ROW,
        height_ratios=[1, 0.1] * NUM_CAMERAS * NUM_LICKS,
    )
    gs.update(wspace=0.0, hspace=0.0)

    video_frame_times = utils.get_video_frame_times(
        session.sync_data, *session.video_paths
    )
    for vid_idx, (video_path, frame_times) in enumerate(video_frame_times.items()):
        if "eye" in video_path.stem.lower():
            continue
        for lick_idx, lick_time in enumerate(lick_times):
            v = utils.get_video_data(video_path)  # TODO open with upath from cloud

            closest_frame_index = np.nanargmin(np.abs(frame_times - lick_time))  # type: ignore[operator]
            frame_indices = np.arange(
                closest_frame_index - FRAMES_EITHER_SIDE_OF_LICK,
                closest_frame_index + FRAMES_EITHER_SIDE_OF_LICK + 1,
            )

            # video frames around lick time
            for frame_idx, frame_index in enumerate(frame_indices):
                v.set(cv2.CAP_PROP_POS_FRAMES, int(frame_index))
                ret, frame = v.read()
                ax = fig.add_subplot(
                    gs[
                        (vid_idx * ROWS_PER_VIDEO) + (lick_idx * ROWS_PER_LICK),
                        frame_idx,
                    ]
                )
                if "beh" in video_path.stem.lower():
                    x = slice(0, frame.shape[1] // 3)
                    y = slice(frame.shape[0] // 5, frame.shape[0] // 2)
                if "face" in video_path.stem.lower():
                    ymid, xmid = frame.shape[0] // 2, frame.shape[1] // 2
                    yspan, xspan = frame.shape[0] // 5, frame.shape[1] // 5
                    x = slice(xmid - xspan, xmid + xspan)
                    y = slice(ymid - yspan, ymid + yspan)
                im = ax.imshow(frame[y, x])
                im.set_clim([0, 255])
                ax.axis("off")
                if frame_index == closest_frame_index:
                    trial = (
                        np.searchsorted(  # type: ignore[call-overload]
                            session.trials[:].start_time, lick_time, "right"
                        )
                        - 1
                    )
                    ax.set_title(f"{trial=}, {lick_time=:.1f} s", fontsize=8)

            # markers for frame and lick times
            ax = fig.add_subplot(
                gs[
                    1 + (vid_idx * ROWS_PER_VIDEO) + (lick_idx * ROWS_PER_LICK),
                    :FRAMES_PER_ROW,
                ]
            )
            linekwargs = {"linewidths": 3}
            ax.eventplot(frame_times[frame_indices], label="frame", **linekwargs)
            ax.eventplot([lick_time], color="red", label="lick", **linekwargs)
            ax.axis("off")
            if lick_idx == vid_idx == 0:
                ax.legend(fontsize=8, fancybox=True, ncol=2, loc="upper right")

    plt.tight_layout()



from __future__ import annotations

import collections.abc
import datetime
import doctest
import functools
import logging
import reprlib
from collections.abc import Generator, Iterator, Sequence
from typing import (
    Any,
    Literal,
)

import numpy as np
import pandas as pd
import polars as pl

import npc_sessions.utils as utils

logger = logging.getLogger(__name__)

EXCLUDE_IF_ONLY_NANS = False
"""If True, exclude columns with all NaN values from the resulting dict."""


class PropertyDict(collections.abc.Mapping):
    """Dict type, where the keys are the class's properties (regular attributes
    and property getters) which don't have a leading underscore, and values are
    corresponding property values.

    Methods can be added and used as normal, they're just not visible in the
    dict.
    """

    @property
    def _properties(self) -> tuple[str, ...]:
        """Names of properties without leading underscores. No methods.
        Excludes properties with all NaN values."""
        self_and_super_attrs = tuple(
            attr
            for cls in reversed(self.__class__.__mro__)
            for attr in cls.__dict__.keys()
        )
        dict_attrs = collections.abc.Mapping.__dict__.keys()
        no_dict_attrs = tuple(
            attr for attr in self_and_super_attrs if attr not in dict_attrs
        )
        no_leading_underscore = tuple(attr for attr in no_dict_attrs if attr[0] != "_")
        no_functions = tuple(
            attr for attr in no_leading_underscore if not callable(getattr(self, attr))
        )

        def all_nan(attr):
            return (
                isinstance(a := getattr(self, attr), np.ndarray)
                and a.dtype == np.floating
                and all(np.isnan(a))
            )

        if EXCLUDE_IF_ONLY_NANS:
            return tuple(attr for attr in no_functions if not all_nan(attr))
        return no_functions

    @property
    def _dict(self) -> dict[str, Any]:
        return {attr: getattr(self, attr) for attr in self._properties}

    def __getitem__(self, key) -> Any:
        return self._dict[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __repr__(self) -> str:
        return reprlib.repr(self._dict)

    def to_add_trial_column(
        self,
    ) -> Generator[dict[Literal["name", "description"], str], None, None]:
        """Name and description for each trial column.

        Does not include `start_time` and `stop_time`: those columns
        always exist in an nwb trials table, trying to add them again raises
        an error.

        Iterate over result and unpack each dict:

        >>> for column in obj.to_add_trial_column(): # doctest: +SKIP
        ...    nwb_file.add_trial_column(**column)

        """
        attrs = self._properties
        descriptions = self._docstrings
        missing = tuple(column for column in attrs if column not in descriptions)
        if any(missing):
            logger.warning(
                f"These properties do not have descriptions (add docstrings to their property getters): {missing}"
            )
        descriptions.update(dict(zip(missing, ("" for _ in missing))))
        descriptions.pop("start_time", None)
        descriptions.pop("stop_time", None)
        return (
            {"name": name, "description": description}
            for name, description in descriptions.items()
        )

    def to_add_trial(
        self,
    ) -> Generator[dict[str, int | float | str | datetime.datetime], None, None]:
        """Column name and value for each trial.

        Iterate over result and unpack each dict:

        >>> for trial in obj.to_add_trial(): # doctest: +SKIP
        ...    nwb_file.add_trial(**trial)

        """
        mandatory_columns = ("start_time", "stop_time")
        if any(mandatory not in self._properties for mandatory in mandatory_columns):
            raise AttributeError(
                f"{self} is missing one of {mandatory_columns = } required for nwb trials table"
            )

        for trial in self.to_dataframe().iterrows():
            yield dict(trial[1])

    def to_dataframe(self) -> pd.DataFrame:
        def get_dtype(attr: str):
            """Get dtype of attribute"""
            # if any(key in attr for key in ('index', 'idx', 'number')):
            #     return 'Int32'
            # if 'time' in attr:
            #     return 'Float32'
            return None

        return pd.DataFrame(
            data={k: pd.Series(v, dtype=get_dtype(k)) for k, v in self.items()}
        )

    @property
    def _df(self) -> pl.DataFrame:
        def get_dtype(attr: str, value: Any) -> type[pl.DataType] | None:
            """Get dtype of attribute"""
            if any(key in attr for key in ("index", "idx", "number")):
                return pl.Int64
            if "time" in attr:
                return pl.Float64
            dtype = getattr(value, "dtype", None)
            if dtype and dtype in (np.str_, str):
                return pl.Utf8
            return None

        return pl.DataFrame(
            pl.Series(k, v, dtype=get_dtype(k, v), nan_to_null=True)
            for k, v in self.items()
        )

    @property
    def _docstrings(self) -> dict[str, str]:
        """Docstrings of property getter methods that have no leading
        underscore in their name.

        - getting the docstring of a regular property/attribute isn't easy:
        if we query its docstring in the same way as a property getter/method,
        we'll just receive the docstring for its value's type.
        """

        def cls_attr(attr):
            return getattr(self.__class__, attr)

        {
            attr: ""
            for attr in self._properties
            if not isinstance(
                cls_attr(attr),
                (property, functools.cached_property, utils.cached_property),
            )
        }
        property_getters = {
            attr: cls_attr(attr).__doc__ or ""
            for attr in self._properties
            if isinstance(
                cls_attr(attr),
                (property, functools.cached_property, utils.cached_property),
            )
        }

        def fmt(docstring):
            return (
                docstring.replace("\n", " ")
                .replace("  ", "")
                .replace(".- ", "; ")
                .replace("- ", "; ")
                .replace(" ; ", "; ")
                .strip(". ")
            )

        return {
            attr: fmt(cls_attr(attr).__doc__)
            if cls_attr(attr).__doc__
            else ""  # if no docstring present, __doc__ is None
            for attr in property_getters
        }


class TestPropertyDict(PropertyDict):
    """
    >>> obj = TestPropertyDict()
    >>> obj
    {'no_docstring': True, 'visible_property': True, 'visible_property_getter': True}

    >>> obj.invisible_method()
    True

    >>> obj._docstrings
    {'visible_property_getter': 'Docstring available', 'no_docstring': ''}
    """

    visible_property = True

    @property
    def visible_property_getter(self):
        """Docstring available"""
        return True

    @property
    def no_docstring(self):
        return True

    _invisible_property: Any = None

    def invisible_method(self):
        """Docstring not available"""
        return True


class TestPropertyDictInheritance(TestPropertyDict):
    """
    >>> obj = TestPropertyDictInheritance()
    >>> obj
    {'no_docstring': True, 'visible_property': True, 'visible_property_getter': True}
    """

    pass


class TestPropertyDictExports(PropertyDict):
    """
    >>> obj = TestPropertyDictExports()

    >>> for kwargs in obj.to_add_trial_column():
    ...     print(kwargs)
    {'name': 'test_start_time', 'description': 'Start of trial'}
    {'name': 'test_stop_time', 'description': 'End of trial'}

    note: `start_time` and `stop_time` cannot be added as columns (they always
    exist)

    >>> for kwargs in obj.to_add_trial():
    ...     print(kwargs)
    {'start_time': 0.0, 'stop_time': 1.0, 'test_start_time': 1.0, 'test_stop_time': 1.5}
    {'start_time': 1.0, 'stop_time': 2.0, 'test_start_time': 2.0, 'test_stop_time': 2.5}
    """

    start_time = [0, 1]
    stop_time = [1, 2]

    @property
    def test_start_time(self) -> Sequence[float]:
        "Start of trial"
        return [1.0, 2.0]

    @property
    def test_stop_time(self) -> Sequence[float]:
        "End of trial"
        return [start + 0.5 for start in self.test_start_time]


if __name__ == "__main__":
    doctest.testmod()
